/* GIO - GLib Input, Output and Streaming Library
 * 
 * Copyright (C) 2006-2007 Red Hat, Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 * Author: Alexander Larsson <alexl@redhat.com>
 */

#include <config.h>

#include <glib.h>
#include <glib/gi18n.h>
#include "daemon-main.h"
#include G_STRINGIFY(BACKEND_HEADER)

int
main (int argc, char *argv[])
{
#ifndef BACKEND_USES_GVFS
  g_setenv ("GIO_USE_VFS", "local", TRUE);
#endif
#ifdef BACKEND_PRE_SETUP_FUNC
  BACKEND_PRE_SETUP_FUNC (&argc, &argv);
#endif
  daemon_init ();
#ifdef BACKEND_SETUP_FUNC
  BACKEND_SETUP_FUNC ();
#else
  daemon_setup ();
#endif
  daemon_main (argc, argv,
#ifdef MAX_JOB_THREADS
	       MAX_JOB_THREADS,
#else
	       -1, 
#endif
#ifdef DEFAULT_BACKEND_TYPE
	       G_STRINGIFY(DEFAULT_BACKEND_TYPE),
#else
	       NULL,
#endif
#ifdef MOUNTABLE_DBUS_NAME
	       G_STRINGIFY(MOUNTABLE_DBUS_NAME),
#else
	       NULL,
#endif
	       BACKEND_TYPES
	       NULL);
	       
  return 0;
}
